"""
Utils used for the CLI scripts.
"""

import inspect
import pathlib
from collections.abc import Callable
from functools import wraps
from typing import Any

import albumentations as A
from rich import box
from rich.console import Console
from rich.pretty import Pretty
from rich.table import Table


def print_variables_as_table(
    c1_title: str, c2_title: str, title: str = "Variables Table", **kwargs: Any
) -> None:
    """
    Prints variables in a formatted table using the rich library.

    Args:
        c1_title (str): Title of the first column.
        c2_title (str): Title of the second column.
        title (str): Title of the table.
        **kwargs (Any): Variable names and values to be printed.
    """
    console = Console()
    console.print("\n")
    table = Table(title=title, show_header=True, header_style="bold blue", box=box.ROUNDED)
    table.add_column(c1_title, min_width=20, justify="left", style="bold")
    table.add_column(c2_title, min_width=60, justify="left", style="bold")

    for key, value in kwargs.items():
        if isinstance(value, pathlib.Path):
            value = str(value)  # noqa: PLW2901
        table.add_row(f"[bold]{key}[/bold]", Pretty(value))

    console.print(table)


def print_params(
    table_title: str = "Parameters Table", c1_title: str = "Variable", c2_title: str = "Value"
) -> Callable:
    """
    A decorator that prints the parameters of a function in a formatted table
    using the rich library.

    Args:
        c1_title (str, optional): Title of the first column. Defaults to "Variable".
        c2_title (str, optional): Title of the second column. Defaults to "Value".
        table_title (str, optional): Title of the table. Defaults to "Parameters Table".

    Returns:
        Callable: The wrapped function with parameter printing functionality.
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            func_signature = inspect.signature(func)
            bound_arguments = func_signature.bind(*args, **kwargs)
            bound_arguments.apply_defaults()
            params = dict(bound_arguments.arguments.items())
            print_variables_as_table(c1_title, c2_title, table_title, **params)
            return func(*args, **kwargs)

        return wrapper

    return decorator


def print_train_details(augmentation: A.Compose, config: dict[str, Any]) -> None:
    console = Console()
    console.print("\n")
    console.print("[bold blue]Augmentation Pipeline:[/bold blue]")
    console.print(Pretty(augmentation))
    console.print("\n")
    console.print("[bold blue]Configuration:[/bold blue]")
    console.print(Pretty(config))
    console.print("\n")



"""
Model definitions for the FastLP OCR.
"""

from typing import Literal

from keras.src.activations import softmax
from keras.src.layers import (
    Activation,
    Concatenate,
    Dense,
    Dropout,
    GlobalAveragePooling2D,
    Input,
    Rescaling,
    Reshape,
    Softmax,
)
from keras.src.models import Model

from fast_plate_ocr.train.model.layer_blocks import (
    block_average_conv_down,
    block_bn,
    block_max_conv_down,
    block_no_activation,
)


def cnn_ocr_model(
    h: int,
    w: int,
    max_plate_slots: int,
    vocabulary_size: int,
    dense: bool = True,
    activation: str = "relu",
    pool_layer: Literal["avg", "max"] = "max",
) -> Model:
    """
    OCR model implemented with just CNN layers (v2).
    """
    input_tensor = Input((h, w, 1))
    x = Rescaling(1.0 / 255)(input_tensor)
    # Pooling-Conv layer
    if pool_layer == "avg":
        block_pool_conv = block_average_conv_down
    elif pool_layer == "max":
        block_pool_conv = block_max_conv_down
    # Backbone
    x = block_pool_conv(x, n_c=32, padding="same", activation=activation)
    x, _ = block_bn(x, k=3, n_c=64, s=1, padding="same", activation=activation)
    x, _ = block_bn(x, k=1, n_c=64, s=1, padding="same", activation=activation)
    x = block_pool_conv(x, n_c=64, padding="same", activation=activation)
    x, _ = block_bn(x, k=3, n_c=128, s=1, padding="same", activation=activation)
    x, _ = block_bn(x, k=1, n_c=128, s=1, padding="same", activation=activation)
    x = block_pool_conv(x, n_c=128, padding="same", activation=activation)
    x, _ = block_bn(x, k=3, n_c=128, s=1, padding="same", activation=activation)
    x, _ = block_bn(x, k=1, n_c=256, s=1, padding="same", activation=activation)
    x = block_pool_conv(x, n_c=256, padding="same", activation=activation)
    x, _ = block_bn(x, k=1, n_c=512, s=1, padding="same", activation=activation)
    x, _ = block_bn(x, k=1, n_c=1024, s=1, padding="same", activation=activation)
    x = (
        head(x, max_plate_slots, vocabulary_size)
        if dense
        else head_no_fc(x, max_plate_slots, vocabulary_size)
    )
    return Model(inputs=input_tensor, outputs=x)


def head(x, max_plate_slots: int, vocabulary_size: int):
    """
    Model's head with Fully Connected (FC) layers.
    """
    x = GlobalAveragePooling2D()(x)
    # dropout for more robust learning
    x = Dropout(0.5)(x)
    dense_outputs = [
        Activation(softmax)(Dense(units=vocabulary_size)(x)) for _ in range(max_plate_slots)
    ]
    # concat all the dense outputs
    x = Concatenate()(dense_outputs)
    return x


def head_no_fc(x, max_plate_slots: int, vocabulary_size: int):
    """
    Model's head without Fully Connected (FC) layers.
    """
    x = block_no_activation(x, k=1, n_c=max_plate_slots * vocabulary_size, s=1, padding="same")
    x = GlobalAveragePooling2D()(x)
    x = Reshape((max_plate_slots, vocabulary_size, 1))(x)
    x = Softmax(axis=-2)(x)
    return x



"""
Common utilities used across the package.
"""

import logging
import time
from collections.abc import Callable, Iterator
from contextlib import contextmanager


@contextmanager
def log_time_taken(process_name: str) -> Iterator[None]:
    """
    A concise context manager to time code snippets and log the result.

    Usage:
    with log_time_taken("process_name"):
        # Code snippet to be timed

    :param process_name: Name of the process being timed.
    """
    time_start: float = time.perf_counter()
    try:
        yield
    finally:
        time_end: float = time.perf_counter()
        time_elapsed: float = time_end - time_start
        logger = logging.getLogger(__name__)
        logger.info("Computation time of '%s' = %.3fms", process_name, 1_000 * time_elapsed)


@contextmanager
def measure_time() -> Iterator[Callable[[], float]]:
    """
    A context manager for measuring execution time (in milliseconds) within its code block.

    usage:
        with code_timer() as timer:
            # Code snippet to be timed
        print(f"Code took: {timer()} seconds")
    """
    start_time = end_time = time.perf_counter()
    yield lambda: (end_time - start_time) * 1_000
    end_time = time.perf_counter()



"""
Utility functions module
"""

import logging
import pathlib
import random

import cv2
import keras
import numpy as np
import numpy.typing as npt

from fast_plate_ocr.train.model.custom import (
    cat_acc_metric,
    cce_loss,
    plate_acc_metric,
    top_3_k_metric,
)


def one_hot_plate(plate: str, alphabet: str) -> list[list[int]]:
    return [[0 if char != letter else 1 for char in alphabet] for letter in plate]


def target_transform(
    plate_text: str,
    max_plate_slots: int,
    alphabet: str,
    pad_char: str,
) -> npt.NDArray[np.uint8]:
    # Pad the plates which length is smaller than 'max_plate_slots'
    plate_text = plate_text.ljust(max_plate_slots, pad_char)
    # Generate numpy arrays with one-hot encoding of plates
    encoded_plate = np.array(one_hot_plate(plate_text, alphabet=alphabet), dtype=np.uint8)
    return encoded_plate


def read_plate_image(image_path: str, img_height: int, img_width: int) -> npt.NDArray:
    """
    Read and resize a license plate image.

    :param image_path: The path to the license plate image.
    :param img_height: The desired height of the resized image.
    :param img_width: The desired width of the resized image.
    :return: The resized license plate image as a NumPy array.
    """
    img = cv2.imread(image_path, cv2.IMREAD_GRAYSCALE)
    img = cv2.resize(img, (img_width, img_height), interpolation=cv2.INTER_LINEAR)
    img = np.expand_dims(img, -1)
    return img


def load_keras_model(
    model_path: str | pathlib.Path,
    vocab_size: int,
    max_plate_slots: int,
) -> keras.Model:
    """
    Utility helper function to load the keras OCR model.
    """
    custom_objects = {
        "cce": cce_loss(vocabulary_size=vocab_size),
        "cat_acc": cat_acc_metric(max_plate_slots=max_plate_slots, vocabulary_size=vocab_size),
        "plate_acc": plate_acc_metric(max_plate_slots=max_plate_slots, vocabulary_size=vocab_size),
        "top_3_k": top_3_k_metric(vocabulary_size=vocab_size),
    }
    model = keras.models.load_model(model_path, custom_objects=custom_objects)
    return model


IMG_EXTENSIONS: set[str] = {".jpg", ".jpeg", ".png", ".bmp", ".gif", ".tiff", ".webp"}
"""Valid image extensions for the scope of this script."""


def load_images_from_folder(
    img_dir: pathlib.Path,
    width: int,
    height: int,
    shuffle: bool = False,
    limit: int | None = None,
) -> list[npt.NDArray]:
    """
    Return all images read from a directory. This uses the same read function used during training.
    """
    image_paths = sorted(
        str(f.resolve()) for f in img_dir.iterdir() if f.is_file() and f.suffix in IMG_EXTENSIONS
    )
    if limit:
        image_paths = image_paths[:limit]
    if shuffle:
        random.shuffle(image_paths)
    images = [read_plate_image(i, img_height=height, img_width=width) for i in image_paths]
    return images


def postprocess_model_output(
    prediction: npt.NDArray,
    alphabet: str,
    max_plate_slots: int,
    vocab_size: int,
) -> tuple[str, npt.NDArray]:
    """
    Return plate text and confidence scores from raw model output.
    """
    prediction = prediction.reshape((max_plate_slots, vocab_size))
    probs = np.max(prediction, axis=-1)
    prediction = np.argmax(prediction, axis=-1)
    plate = "".join([alphabet[x] for x in prediction])
    return plate, probs


def low_confidence_positions(probs, thresh=0.3) -> npt.NDArray:
    """Returns indices of elements in `probs` less than `thresh`, indicating low confidence."""
    return np.where(np.array(probs) < thresh)[0]


def display_predictions(
    image: npt.NDArray,
    plate: str,
    probs: npt.NDArray,
    low_conf_thresh: float,
) -> None:
    """
    Display plate and corresponding prediction.
    """
    plate_str = "".join(plate)
    logging.info("Plate: %s", plate_str)
    logging.info("Confidence: %s", probs)
    image_to_show = cv2.resize(image, None, fx=3, fy=3, interpolation=cv2.INTER_LINEAR)
    # Converting to BGR for color text
    image_to_show = cv2.cvtColor(image_to_show, cv2.COLOR_GRAY2RGB)
    # Average probabilities
    avg_prob = np.mean(probs) * 100
    cv2.putText(
        image_to_show,
        f"{plate_str}  {avg_prob:.{2}f}%",
        org=(5, 30),
        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
        fontScale=1,
        color=(0, 0, 0),
        lineType=1,
        thickness=6,
    )
    cv2.putText(
        image_to_show,
        f"{plate_str}  {avg_prob:.{2}f}%",
        org=(5, 30),
        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
        fontScale=1,
        color=(255, 255, 255),
        lineType=1,
        thickness=2,
    )
    # Display character with low confidence
    low_conf_chars = "Low conf. on: " + " ".join(
        [plate[i] for i in low_confidence_positions(probs, thresh=low_conf_thresh)]
    )
    cv2.putText(
        image_to_show,
        low_conf_chars,
        org=(5, 200),
        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
        fontScale=0.7,
        color=(0, 0, 220),
        lineType=1,
        thickness=2,
    )
    cv2.imshow("plates", image_to_show)
    if cv2.waitKey(0) & 0xFF == ord("q"):
        return



"""
Script for converting Keras models to ONNX format.
"""

import logging
import pathlib
import shutil
from tempfile import NamedTemporaryFile

import click
import numpy as np
import onnx
import onnxruntime as rt
import onnxsim
import tensorflow as tf
import tf2onnx
from tf2onnx import constants as tf2onnx_constants

from fast_plate_ocr.common.utils import log_time_taken
from fast_plate_ocr.train.model.config import load_config_from_yaml
from fast_plate_ocr.train.utilities.utils import load_keras_model

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
)


# pylint: disable=too-many-arguments,too-many-locals


@click.command(context_settings={"max_content_width": 120})
@click.option(
    "-m",
    "--model",
    "model_path",
    required=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, path_type=pathlib.Path),
    help="Path to the saved .keras model.",
)
@click.option(
    "--output-path",
    required=True,
    type=str,
    help="Output name for ONNX model.",
)
@click.option(
    "--simplify/--no-simplify",
    default=False,
    show_default=True,
    help="Simplify ONNX model using onnxsim.",
)
@click.option(
    "--config-file",
    required=True,
    type=click.Path(exists=True, file_okay=True, path_type=pathlib.Path),
    help="Path pointing to the model license plate OCR config.",
)
@click.option(
    "--opset",
    default=16,
    type=click.IntRange(max=max(tf2onnx_constants.OPSET_TO_IR_VERSION)),
    show_default=True,
    help="Opset version for ONNX.",
)
def export_onnx(
    model_path: pathlib.Path,
    output_path: str,
    simplify: bool,
    config_file: pathlib.Path,
    opset: int,
) -> None:
    """
    Export Keras models to ONNX format.
    """
    config = load_config_from_yaml(config_file)
    model = load_keras_model(
        model_path,
        vocab_size=config.vocabulary_size,
        max_plate_slots=config.max_plate_slots,
    )
    spec = (tf.TensorSpec((None, config.img_height, config.img_width, 1), tf.uint8, name="input"),)
    # Convert from Keras to ONNX using tf2onnx library
    with NamedTemporaryFile(suffix=".onnx") as tmp:
        tmp_onnx = tmp.name
        model_proto, _ = tf2onnx.convert.from_keras(
            model,
            input_signature=spec,
            opset=opset,
            output_path=tmp_onnx,
        )
        if simplify:
            logging.info("Simplifying ONNX model ...")
            model_simp, check = onnxsim.simplify(onnx.load(tmp_onnx))
            assert check, "Simplified ONNX model could not be validated!"
            onnx.save(model_simp, output_path)
        else:
            shutil.copy(tmp_onnx, output_path)
    output_names = [n.name for n in model_proto.graph.output]
    x = np.random.randint(0, 256, size=(1, config.img_height, config.img_width, 1), dtype=np.uint8)
    # Run dummy inference and log time taken
    m = rt.InferenceSession(output_path)
    with log_time_taken("ONNX inference took:"):
        onnx_pred = m.run(output_names, {"input": x})
    # Check if ONNX and keras have the same results
    if not np.allclose(model.predict(x, verbose=0), onnx_pred[0], rtol=1e-5, atol=1e-5):
        logging.warning("ONNX model output was not close to Keras model for the given tolerance!")
    logging.info("Model converted to ONNX! Saved at %s", output_path)


if __name__ == "__main__":
    export_onnx()



"""
ONNX inference module.
"""

import logging
import os
import pathlib
from collections.abc import Sequence
from typing import Literal

import numpy as np
import numpy.typing as npt
import onnxruntime as ort
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from fast_plate_ocr.common.utils import measure_time
from fast_plate_ocr.inference import hub
from fast_plate_ocr.inference.config import load_config_from_yaml
from fast_plate_ocr.inference.hub import OcrModel
from fast_plate_ocr.inference.process import postprocess_output, preprocess_image, read_plate_image


def _load_image_from_source(
    source: str | list[str] | npt.NDArray | list[npt.NDArray],
) -> npt.NDArray | list[npt.NDArray]:
    """
    Loads an image from a given source.

    :param source: Path to the input image file, list of paths, or numpy array representing one or
     multiple images.
    :return: Numpy array representing the input image(s) or a list of numpy arrays.
    """
    if isinstance(source, str):
        # Shape returned (H, W)
        return read_plate_image(source)

    if isinstance(source, list):
        # Are image paths
        if all(isinstance(s, str) for s in source):
            # List returned with array item of shape (H, W)
            return [read_plate_image(i) for i in source]  # type: ignore[arg-type]
        # Are list of numpy arrays
        if all(isinstance(a, np.ndarray) for a in source):
            # List returned with array item of shape (H, W)
            return source  # type: ignore[return-value]
        raise ValueError("Expected source to be a list of `str` or `np.ndarray`!")

    if isinstance(source, np.ndarray):
        # Squeeze grayscale channel dimension if supplied
        source = source.squeeze()
        if source.ndim != 2:
            raise ValueError("Expected source array to be of shape (H, W) or (H, W, 1).")
        # Shape returned (H, W)
        return source

    raise ValueError("Unsupported input type. Only file path or numpy array is supported.")


class ONNXPlateRecognizer:
    """
    ONNX inference class for performing license plates OCR.
    """

    def __init__(
        self,
        hub_ocr_model: OcrModel | None = None,
        device: Literal["cuda", "cpu", "auto"] = "auto",
        providers: Sequence[str | tuple[str, dict]] | None = None,
        sess_options: ort.SessionOptions | None = None,
        model_path: str | os.PathLike[str] | None = None,
        config_path: str | os.PathLike[str] | None = None,
        force_download: bool = False,
    ) -> None:
        """
        Initializes the ONNXPlateRecognizer with the specified OCR model and inference device.

        The current OCR models available from the HUB are:

        - `argentinian-plates-cnn-model`: OCR for Argentinian license plates. Uses fully conv
            architecture.
        - `argentinian-plates-cnn-synth-model`: OCR for Argentinian license plates trained with
            synthetic and real data. Uses fully conv architecture.
        - `european-plates-mobile-vit-v2-model`: OCR for European license plates. Uses MobileVIT-2
            for the backbone.
        - `global-plates-mobile-vit-v2-model`: OCR for Global license plates (+65 countries).
            Uses MobileVIT-2 for the backbone.

        Args:
            hub_ocr_model: Name of the OCR model to use from the HUB.
            device: Device type for inference. Should be one of ('cpu', 'cuda', 'auto'). If
                'auto' mode, the device will be deduced from
                `onnxruntime.get_available_providers()`.
            providers: Optional sequence of providers in order of decreasing precedence. If not
                specified, all available providers are used based on the device argument.
            sess_options: Advanced session options for ONNX Runtime.
            model_path: Path to ONNX model file to use (In case you want to use a custom one).
            config_path: Path to config file to use (In case you want to use a custom one).
            force_download: Force and download the model, even if it already exists.
        Returns:
            None.
        """
        self.logger = logging.getLogger(__name__)

        if providers is not None:
            self.providers = providers
            self.logger.info("Using custom providers: %s", providers)
        else:
            if device == "cuda":
                self.providers = ["CUDAExecutionProvider"]
            elif device == "cpu":
                self.providers = ["CPUExecutionProvider"]
            elif device == "auto":
                self.providers = ort.get_available_providers()
            else:
                raise ValueError(
                    f"Device should be one of ('cpu', 'cuda', 'auto'). Got '{device}'."
                )

            self.logger.info("Using device '%s' with providers: %s", device, self.providers)

        if model_path and config_path:
            model_path = pathlib.Path(model_path)
            config_path = pathlib.Path(config_path)
            if not model_path.exists() or not config_path.exists():
                raise FileNotFoundError("Missing model/config file!")
            self.model_name = model_path.stem
        elif hub_ocr_model:
            self.model_name = hub_ocr_model
            model_path, config_path = hub.download_model(
                model_name=hub_ocr_model, force_download=force_download
            )
        else:
            raise ValueError(
                "Either provide a model from the HUB or a custom model_path and config_path"
            )

        self.config = load_config_from_yaml(config_path)
        self.model = ort.InferenceSession(
            model_path, providers=self.providers, sess_options=sess_options
        )
        self.logger.info("Using ONNX Runtime with %s.", self.providers)

    def benchmark(self, n_iter: int = 10_000, include_processing: bool = False) -> None:
        """
        Benchmark time taken to run the OCR model. This reports the average inference time and the
        throughput in plates per second.

        Args:
            n_iter: The number of iterations to run the benchmark. This determines how many times
                the inference will be executed to compute the average performance metrics.
            include_processing: Indicates whether the benchmark should include preprocessing and
                postprocessing times in the measurement.
        """
        cum_time = 0.0
        x = np.random.randint(
            0, 256, size=(1, self.config["img_height"], self.config["img_width"], 1), dtype=np.uint8
        )
        for _ in range(n_iter):
            with measure_time() as time_taken:
                if include_processing:
                    self.run(x)
                else:
                    self.model.run(None, {"input": x})
            cum_time += time_taken()

        avg_time = (cum_time / n_iter) if n_iter > 0 else 0.0
        avg_pps = (1_000 / avg_time) if n_iter > 0 else 0.0

        console = Console()
        model_info = Panel(
            Text(f"Model: {self.model_name}\nProviders: {self.providers}", style="bold green"),
            title="Model Information",
            border_style="bright_blue",
            expand=False,
        )
        console.print(model_info)
        table = Table(title=f"Benchmark for '{self.model_name}' Model", border_style="bright_blue")
        table.add_column("Metric", justify="center", style="cyan", no_wrap=True)
        table.add_column("Value", justify="center", style="magenta")
        table.add_row("Number of Iterations", str(n_iter))
        table.add_row("Average Time (ms)", f"{avg_time:.4f}")
        table.add_row("Plates Per Second (PPS)", f"{avg_pps:.4f}")
        console.print(table)

    def run(
        self,
        source: str | list[str] | npt.NDArray | list[npt.NDArray],
        return_confidence: bool = False,
    ) -> tuple[list[str], npt.NDArray] | list[str]:
        """
        Performs OCR to recognize license plate characters from an image or a list of images.

        Args:
            source: The path(s) to the image(s), a numpy array representing an image or a list
                of NumPy arrays. If a numpy array is provided, it is expected to already be in
                grayscale format, with shape `(H, W) `or `(H, W, 1)`. A list of numpy arrays with
                different image sizes may also be provided.
            return_confidence: Whether to return confidence scores along with plate predictions.

        Returns:
            A list of plates for each input image. If `return_confidence` is True, a numpy
                array is returned with the shape `(N, plate_slots)`, where N is the batch size and
                each plate slot is the confidence for the recognized license plate character.
        """
        x = _load_image_from_source(source)
        # Preprocess
        x = preprocess_image(x, self.config["img_height"], self.config["img_width"])
        # Run model
        y: list[npt.NDArray] = self.model.run(None, {"input": x})
        # Postprocess model output
        return postprocess_output(
            y[0],
            self.config["max_plate_slots"],
            self.config["alphabet"],
            return_confidence=return_confidence,
        )



"""
Utility functions for processing model input/output.
"""

import os

import cv2
import numpy as np
import numpy.typing as npt


def read_plate_image(image_path: str) -> npt.NDArray:
    """
    Read image from disk as a grayscale image.

    :param image_path: The path to the license plate image.
    :return: The image as a NumPy array.
    """
    if not os.path.exists(image_path):
        raise ValueError(f"{image_path} file doesn't exist!")
    return cv2.imread(image_path, cv2.IMREAD_GRAYSCALE)


def preprocess_image(
    image: npt.NDArray | list[npt.NDArray], img_height: int, img_width: int
) -> npt.NDArray:
    """
    Preprocess the image(s), so they're ready to be fed to the model.

    Note: We don't normalize the pixel values between [0, 1] here, because that the model first
    layer does that.

    :param image: The image(s) contained in a NumPy array.
    :param img_height: The desired height of the resized image.
    :param img_width: The desired width of the resized image.
    :return: A numpy array with shape (N, H, W, 1).
    """
    # Add batch dimension: (H, W) -> (1, H, W)
    if isinstance(image, np.ndarray):
        image = np.expand_dims(image, axis=0)

    imgs = np.array(
        [
            cv2.resize(im.squeeze(), (img_width, img_height), interpolation=cv2.INTER_LINEAR)
            for im in image
        ]
    )
    # Add channel dimension
    imgs = np.expand_dims(imgs, axis=-1)
    return imgs


def postprocess_output(
    model_output: npt.NDArray,
    max_plate_slots: int,
    model_alphabet: str,
    return_confidence: bool = False,
) -> tuple[list[str], npt.NDArray] | list[str]:
    """
    Post-processes model output and return license plate string, and optionally the probabilities.

    :param model_output: Output from the model containing predictions.
    :param max_plate_slots: Maximum number of characters in a license plate.
    :param model_alphabet: Alphabet used by the model for character encoding.
    :param return_confidence: Flag to indicate whether to return confidence scores along with plate
     predictions.
    :return: Decoded license plate characters as a list, optionally with confidence scores. The
     confidence scores have shape (N, max_plate_slots) where N is the batch size.
    """
    predictions = model_output.reshape((-1, max_plate_slots, len(model_alphabet)))
    prediction_indices = np.argmax(predictions, axis=-1)
    alphabet_array = np.array(list(model_alphabet))
    plate_chars = alphabet_array[prediction_indices]
    plates: list[str] = np.apply_along_axis("".join, 1, plate_chars).tolist()
    if return_confidence:
        probs = np.max(predictions, axis=-1)
        return plates, probs
    return plates



"""
Utilities used around the inference package.
"""

import os
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import IO, Any


@contextmanager
def safe_write(
    file: str | os.PathLike[str],
    mode: str = "wb",
    encoding: str | None = None,
    **kwargs: Any,
) -> Iterator[IO]:
    """
    Context manager for safe file writing.

    Opens the specified file for writing and yields a file object.
    If an exception occurs during writing, the file is removed before raising the exception.
    """
    try:
        with open(file, mode, encoding=encoding, **kwargs) as f:
            yield f
    except Exception as e:
        Path(file).unlink(missing_ok=True)
        raise e



"""
Script for validating trained OCR models.
"""

import pathlib

import click
from torch.utils.data import DataLoader

from fast_plate_ocr.train.data.dataset import LicensePlateDataset

# Custom metris / losses
from fast_plate_ocr.train.model.config import load_config_from_yaml
from fast_plate_ocr.train.utilities import utils


@click.command(context_settings={"max_content_width": 120})
@click.option(
    "-m",
    "--model",
    "model_path",
    required=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, path_type=pathlib.Path),
    help="Path to the saved .keras model.",
)
@click.option(
    "--config-file",
    required=True,
    type=click.Path(exists=True, file_okay=True, path_type=pathlib.Path),
    help="Path pointing to the model license plate OCR config.",
)
@click.option(
    "-a",
    "--annotations",
    required=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, path_type=pathlib.Path),
    help="Annotations file used for validation.",
)
@click.option(
    "-b",
    "--batch-size",
    default=1,
    show_default=True,
    type=int,
    help="Batch size.",
)
def valid(
    model_path: pathlib.Path,
    config_file: pathlib.Path,
    annotations: pathlib.Path,
    batch_size: int,
) -> None:
    """
    Validate the trained OCR model on a labeled set.
    """
    config = load_config_from_yaml(config_file)
    model = utils.load_keras_model(
        model_path, vocab_size=config.vocabulary_size, max_plate_slots=config.max_plate_slots
    )
    val_torch_dataset = LicensePlateDataset(annotations_file=annotations, config=config)
    val_dataloader = DataLoader(val_torch_dataset, batch_size=batch_size, shuffle=False)
    model.evaluate(val_dataloader)


if __name__ == "__main__":
    valid()



"""
Config values used throughout the code.
"""

from os import PathLike

import yaml
from pydantic import BaseModel, computed_field, model_validator


class PlateOCRConfig(BaseModel, extra="forbid", frozen=True):
    """
    Model License Plate OCR config.
    """

    max_plate_slots: int
    """
    Max number of plate slots supported. This represents the number of model classification heads.
    """

    alphabet: str
    """
    All the possible character set for the model output.
    """
    pad_char: str
    """
    Padding character for plates which length is smaller than MAX_PLATE_SLOTS.
    """
    img_height: int
    """
    Image height which is fed to the model.
    """
    img_width: int
    """
    Image width which is fed to the model.
    """

    @computed_field  # type: ignore[misc]
    @property
    def vocabulary_size(self) -> int:
        return len(self.alphabet)

    @model_validator(mode="after")
    def check_pad_in_alphabet(self) -> "PlateOCRConfig":
        if self.pad_char not in self.alphabet:
            raise ValueError("Pad character must be present in model alphabet.")
        return self


def load_config_from_yaml(yaml_file_path: str | PathLike[str]) -> PlateOCRConfig:
    """Read and parse a yaml containing the Plate OCR config."""
    with open(yaml_file_path, encoding="utf-8") as f_in:
        yaml_content = yaml.safe_load(f_in)
    config = PlateOCRConfig(**yaml_content)
    return config



"""
Main CLI used when training a FastPlateOCR model.
"""

try:
    import click

    from fast_plate_ocr.cli.onnx_converter import export_onnx
    from fast_plate_ocr.cli.train import train
    from fast_plate_ocr.cli.valid import valid
    from fast_plate_ocr.cli.visualize_augmentation import visualize_augmentation
    from fast_plate_ocr.cli.visualize_predictions import visualize_predictions
except ImportError as e:
    raise ImportError("Make sure to 'pip install fast-plate-ocr[train]' to run this!") from e


@click.group(context_settings={"max_content_width": 120})
def main_cli():
    """FastPlateOCR CLI."""


main_cli.add_command(visualize_predictions)
main_cli.add_command(visualize_augmentation)
main_cli.add_command(valid)
main_cli.add_command(train)
main_cli.add_command(export_onnx)



"""
Custom metrics and loss functions.
"""

from keras import losses, metrics, ops


def cat_acc_metric(max_plate_slots: int, vocabulary_size: int):
    """
    Categorical accuracy metric.
    """

    def cat_acc(y_true, y_pred):
        """
        This is simply the CategoricalAccuracy for multi-class label problems. Example if the
        correct label is ABC123 and ABC133 is predicted, it will not give a precision of 0% like
        plate_acc (not completely classified correctly), but 83.3% (5/6).
        """
        y_true = ops.reshape(y_true, newshape=(-1, max_plate_slots, vocabulary_size))
        y_pred = ops.reshape(y_pred, newshape=(-1, max_plate_slots, vocabulary_size))
        return ops.mean(metrics.categorical_accuracy(y_true, y_pred))

    return cat_acc


def plate_acc_metric(max_plate_slots: int, vocabulary_size: int):
    """
    Plate accuracy metric.
    """

    def plate_acc(y_true, y_pred):
        """
        Compute how many plates were correctly classified. For a single plate, if ground truth is
        'ABC 123', and the prediction is 'ABC 123', then this would give a score of 1. If the
        prediction was ABD 123, it would score 0.
        """
        y_true = ops.reshape(y_true, newshape=(-1, max_plate_slots, vocabulary_size))
        y_pred = ops.reshape(y_pred, newshape=(-1, max_plate_slots, vocabulary_size))
        et = ops.equal(ops.argmax(y_true, axis=-1), ops.argmax(y_pred, axis=-1))
        return ops.mean(ops.cast(ops.all(et, axis=-1, keepdims=False), dtype="float32"))

    return plate_acc


def top_3_k_metric(vocabulary_size: int):
    """
    Top 3 K categorical accuracy metric.
    """

    def top_3_k(y_true, y_pred):
        """
        Calculates how often the true character is found in the 3 predictions with the highest
        probability.
        """
        # Reshape into 2-d
        y_true = ops.reshape(y_true, newshape=(-1, vocabulary_size))
        y_pred = ops.reshape(y_pred, newshape=(-1, vocabulary_size))
        return ops.mean(metrics.top_k_categorical_accuracy(y_true, y_pred, k=3))

    return top_3_k


# Custom loss
def cce_loss(vocabulary_size: int, label_smoothing: float = 0.2):
    """
    Categorical cross-entropy loss.
    """

    def cce(y_true, y_pred):
        """
        Computes the categorical cross-entropy loss.
        """
        y_true = ops.reshape(y_true, newshape=(-1, vocabulary_size))
        y_pred = ops.reshape(y_pred, newshape=(-1, vocabulary_size))
        return ops.mean(
            losses.categorical_crossentropy(
                y_true, y_pred, from_logits=False, label_smoothing=label_smoothing
            )
        )

    return cce



"""
Dataset module.
"""

import os
from os import PathLike

import albumentations as A
import numpy.typing as npt
import pandas as pd
from torch.utils.data import Dataset

from fast_plate_ocr.train.model.config import PlateOCRConfig
from fast_plate_ocr.train.utilities import utils


class LicensePlateDataset(Dataset):
    def __init__(
        self,
        annotations_file: str | PathLike[str],
        config: PlateOCRConfig,
        transform: A.Compose | None = None,
    ) -> None:
        annotations = pd.read_csv(annotations_file)
        annotations["image_path"] = (
            os.path.dirname(os.path.realpath(annotations_file)) + os.sep + annotations["image_path"]
        )
        assert (
            annotations["plate_text"].str.len() <= config.max_plate_slots
        ).all(), "Plates are longer than max_plate_slots specified param. Change the parameter."
        self.annotations = annotations.to_numpy()
        self.config = config
        self.transform = transform

    def __len__(self) -> int:
        return self.annotations.shape[0]

    def __getitem__(self, idx) -> tuple[npt.NDArray, npt.NDArray]:
        image_path, plate_text = self.annotations[idx]
        x = utils.read_plate_image(
            image_path=image_path,
            img_height=self.config.img_height,
            img_width=self.config.img_width,
        )
        y = utils.target_transform(
            plate_text=plate_text,
            max_plate_slots=self.config.max_plate_slots,
            alphabet=self.config.alphabet,
            pad_char=self.config.pad_char,
        )
        if self.transform:
            x = self.transform(image=x)["image"]
        return x, y



"""
Script for training the License Plate OCR models.
"""

import pathlib
import shutil
from datetime import datetime
from typing import Literal

import albumentations as A
import click
from keras.src.callbacks import EarlyStopping, ModelCheckpoint, ReduceLROnPlateau, TensorBoard
from keras.src.optimizers import Adam
from torch.utils.data import DataLoader

from fast_plate_ocr.cli.utils import print_params, print_train_details
from fast_plate_ocr.train.data.augmentation import TRAIN_AUGMENTATION
from fast_plate_ocr.train.data.dataset import LicensePlateDataset
from fast_plate_ocr.train.model.config import load_config_from_yaml
from fast_plate_ocr.train.model.custom import (
    cat_acc_metric,
    cce_loss,
    plate_acc_metric,
    top_3_k_metric,
)
from fast_plate_ocr.train.model.models import cnn_ocr_model

# ruff: noqa: PLR0913
# pylint: disable=too-many-arguments,too-many-locals


@click.command(context_settings={"max_content_width": 120})
@click.option(
    "--dense/--no-dense",
    default=True,
    show_default=True,
    help="Whether to use Fully Connected layers in model head or not.",
)
@click.option(
    "--config-file",
    required=True,
    type=click.Path(exists=True, file_okay=True, path_type=pathlib.Path),
    help="Path pointing to the model license plate OCR config.",
)
@click.option(
    "--annotations",
    required=True,
    type=click.Path(exists=True, file_okay=True, path_type=pathlib.Path),
    help="Path pointing to the train annotations CSV file.",
)
@click.option(
    "--val-annotations",
    required=True,
    type=click.Path(exists=True, file_okay=True, path_type=pathlib.Path),
    help="Path pointing to the train validation CSV file.",
)
@click.option(
    "--augmentation-path",
    type=click.Path(exists=True, file_okay=True, path_type=pathlib.Path),
    help="YAML file pointing to the augmentation pipeline saved with Albumentations.save(...)",
)
@click.option(
    "--lr",
    default=1e-3,
    show_default=True,
    type=float,
    help="Initial learning rate to use.",
)
@click.option(
    "--label-smoothing",
    default=0.05,
    show_default=True,
    type=float,
    help="Amount of label smoothing to apply.",
)
@click.option(
    "--batch-size",
    default=128,
    show_default=True,
    type=int,
    help="Batch size for training.",
)
@click.option(
    "--num-workers",
    default=0,
    show_default=True,
    type=int,
    help="How many subprocesses to load data, used in the torch DataLoader.",
)
@click.option(
    "--output-dir",
    default="./trained_models",
    type=click.Path(dir_okay=True, path_type=pathlib.Path),
    help="Output directory where model will be saved.",
)
@click.option(
    "--epochs",
    default=500,
    show_default=True,
    type=int,
    help="Number of training epochs.",
)
@click.option(
    "--tensorboard",
    "-t",
    is_flag=True,
    help="Whether to use TensorBoard visualization tool.",
)
@click.option(
    "--tensorboard-dir",
    "-l",
    default="tensorboard_logs",
    show_default=True,
    type=click.Path(path_type=pathlib.Path),
    help="The path of the directory where to save the TensorBoard log files.",
)
@click.option(
    "--early-stopping-patience",
    default=100,
    show_default=True,
    type=int,
    help="Stop training when 'val_plate_acc' doesn't improve for X epochs.",
)
@click.option(
    "--reduce-lr-patience",
    default=60,
    show_default=True,
    type=int,
    help="Patience to reduce the learning rate if 'val_plate_acc' doesn't improve within X epochs.",
)
@click.option(
    "--reduce-lr-factor",
    default=0.85,
    show_default=True,
    type=float,
    help="Reduce the learning rate by this factor when 'val_plate_acc' doesn't improve.",
)
@click.option(
    "--activation",
    default="relu",
    show_default=True,
    type=str,
    help="Activation function to use.",
)
@click.option(
    "--pool-layer",
    default="max",
    show_default=True,
    type=click.Choice(["max", "avg"]),
    help="Choose the pooling layer to use.",
)
@click.option(
    "--weights-path",
    type=click.Path(exists=True, file_okay=True, path_type=pathlib.Path),
    help="Path to the pretrained model weights file.",
)
@print_params(table_title="CLI Training Parameters", c1_title="Parameter", c2_title="Details")
def train(
    dense: bool,
    config_file: pathlib.Path,
    annotations: pathlib.Path,
    val_annotations: pathlib.Path,
    augmentation_path: pathlib.Path | None,
    lr: float,
    label_smoothing: float,
    batch_size: int,
    num_workers: int,
    output_dir: pathlib.Path,
    epochs: int,
    tensorboard: bool,
    tensorboard_dir: pathlib.Path,
    early_stopping_patience: int,
    reduce_lr_patience: int,
    reduce_lr_factor: float,
    activation: str,
    pool_layer: Literal["max", "avg"],
    weights_path: pathlib.Path | None,
) -> None:
    """
    Train the License Plate OCR model.
    """
    train_augmentation = (
        A.load(augmentation_path, data_format="yaml") if augmentation_path else TRAIN_AUGMENTATION
    )
    config = load_config_from_yaml(config_file)
    print_train_details(train_augmentation, config.model_dump())
    train_torch_dataset = LicensePlateDataset(
        annotations_file=annotations,
        transform=train_augmentation,
        config=config,
    )
    train_dataloader = DataLoader(
        train_torch_dataset, batch_size=batch_size, num_workers=num_workers, shuffle=True
    )

    if val_annotations:
        val_torch_dataset = LicensePlateDataset(
            annotations_file=val_annotations,
            config=config,
        )
        val_dataloader = DataLoader(
            val_torch_dataset, batch_size=batch_size, num_workers=num_workers, shuffle=False
        )
    else:
        val_dataloader = None

    # Train
    model = cnn_ocr_model(
        h=config.img_height,
        w=config.img_width,
        dense=dense,
        max_plate_slots=config.max_plate_slots,
        vocabulary_size=config.vocabulary_size,
        activation=activation,
        pool_layer=pool_layer,
    )

    if weights_path:
        model.load_weights(weights_path)

    model.compile(
        loss=cce_loss(vocabulary_size=config.vocabulary_size, label_smoothing=label_smoothing),
        optimizer=Adam(lr),
        metrics=[
            cat_acc_metric(
                max_plate_slots=config.max_plate_slots, vocabulary_size=config.vocabulary_size
            ),
            plate_acc_metric(
                max_plate_slots=config.max_plate_slots, vocabulary_size=config.vocabulary_size
            ),
            top_3_k_metric(vocabulary_size=config.vocabulary_size),
        ],
    )

    output_dir /= datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    output_dir.mkdir(parents=True, exist_ok=True)
    model_file_path = output_dir / "cnn_ocr-epoch_{epoch:02d}-acc_{val_plate_acc:.3f}.keras"

    # Save params and config used for training
    shutil.copy(config_file, output_dir / "config.yaml")
    A.save(train_augmentation, output_dir / "train_augmentation.yaml", "yaml")

    callbacks = [
        # Reduce the learning rate by 0.5x if 'val_plate_acc' doesn't improve within X epochs
        ReduceLROnPlateau(
            "val_plate_acc",
            patience=reduce_lr_patience,
            factor=reduce_lr_factor,
            min_lr=1e-6,
            verbose=1,
        ),
        # Stop training when 'val_plate_acc' doesn't improve for X epochs
        EarlyStopping(
            monitor="val_plate_acc",
            patience=early_stopping_patience,
            mode="max",
            restore_best_weights=False,
            verbose=1,
        ),
        # We don't use EarlyStopping restore_best_weights=True because it won't restore the best
        # weights when it didn't manage to EarlyStop but finished all epochs
        ModelCheckpoint(
            model_file_path,
            monitor="val_plate_acc",
            mode="max",
            save_best_only=True,
            verbose=1,
        ),
    ]

    if tensorboard:
        run_dir = tensorboard_dir / datetime.now().strftime("run_%Y%m%d_%H%M%S")
        run_dir.mkdir(parents=True, exist_ok=True)
        callbacks.append(TensorBoard(log_dir=run_dir))

    model.fit(train_dataloader, epochs=epochs, validation_data=val_dataloader, callbacks=callbacks)


if __name__ == "__main__":
    train()



"""
Model config reading/parsing for doing inference.
"""

from os import PathLike
from typing import TypedDict

import yaml

# pylint: disable=duplicate-code


class PlateOCRConfig(TypedDict):
    """
    Plate OCR Config used for inference.

    This has the same attributes as the one used in the training Pydantic BaseModel. We use this to
    avoid having Pydantic as a required dependency of the minimal package install.
    """

    max_plate_slots: int
    """
    Max number of plate slots supported. This represents the number of model classification heads.
    """
    alphabet: str
    """
    All the possible character set for the model output.
    """
    pad_char: str
    """
    Padding character for plates which length is smaller than MAX_PLATE_SLOTS.
    """
    img_height: int
    """
    Image height which is fed to the model.
    """
    img_width: int
    """
    Image width which is fed to the model.
    """


def load_config_from_yaml(yaml_file_path: str | PathLike[str]) -> PlateOCRConfig:
    """
    Read and parse a yaml containing the Plate OCR config.

    Note: This is currently not using Pydantic for parsing/validating to avoid adding it a python
    dependency as part of the minimal installation.
    """
    with open(yaml_file_path, encoding="utf-8") as f_in:
        config: PlateOCRConfig = yaml.safe_load(f_in)
    return config



"""
Utilities function used for doing inference with the OCR models.
"""

import logging
import pathlib
import shutil
import urllib.request
from http import HTTPStatus
from typing import Literal

from tqdm.asyncio import tqdm

from fast_plate_ocr.inference.utils import safe_write

BASE_URL: str = "https://github.com/ankandrew/cnn-ocr-lp/releases/download"
OcrModel = Literal[
    "argentinian-plates-cnn-model",
    "argentinian-plates-cnn-synth-model",
    "european-plates-mobile-vit-v2-model",
    "global-plates-mobile-vit-v2-model",
]


AVAILABLE_ONNX_MODELS: dict[OcrModel, tuple[str, str]] = {
    "argentinian-plates-cnn-model": (
        f"{BASE_URL}/arg-plates/arg_cnn_ocr.onnx",
        f"{BASE_URL}/arg-plates/arg_cnn_ocr_config.yaml",
    ),
    "argentinian-plates-cnn-synth-model": (
        f"{BASE_URL}/arg-plates/arg_cnn_ocr_synth.onnx",
        f"{BASE_URL}/arg-plates/arg_cnn_ocr_config.yaml",
    ),
    "european-plates-mobile-vit-v2-model": (
        f"{BASE_URL}/arg-plates/european_mobile_vit_v2_ocr.onnx",
        f"{BASE_URL}/arg-plates/european_mobile_vit_v2_ocr_config.yaml",
    ),
    "global-plates-mobile-vit-v2-model": (
        f"{BASE_URL}/arg-plates/global_mobile_vit_v2_ocr.onnx",
        f"{BASE_URL}/arg-plates/global_mobile_vit_v2_ocr_config.yaml",
    ),
}
"""Available ONNX models for doing inference."""

MODEL_CACHE_DIR: pathlib.Path = pathlib.Path.home() / ".cache" / "fast-plate-ocr"
"""Default location where models will be stored."""


def _download_with_progress(url: str, filename: pathlib.Path) -> None:
    """
    Download utility function with progress bar.

    :param url: URL of the model to download.
    :param filename: Where to save the OCR model.
    """
    with urllib.request.urlopen(url) as response, safe_write(filename, mode="wb") as out_file:
        if response.getcode() != HTTPStatus.OK:
            raise ValueError(f"Failed to download file from {url}. Status code: {response.status}")

        file_size = int(response.headers.get("Content-Length", 0))
        desc = f"Downloading {filename.name}"

        with tqdm.wrapattr(out_file, "write", total=file_size, desc=desc) as f_out:
            shutil.copyfileobj(response, f_out)


def download_model(
    model_name: OcrModel,
    save_directory: pathlib.Path | None = None,
    force_download: bool = False,
) -> tuple[pathlib.Path, pathlib.Path]:
    """
    Download an OCR model and the config to a given directory.

    :param model_name: Which model to download.
    :param save_directory: Directory to save the OCR model. It should point to a folder. If not
     supplied, this will point to '~/.cache/<model_name>'
    :param force_download: Force and download the model if it already exists in `save_directory`.
    :return: A tuple consisting of (model_downloaded_path, config_downloaded_path).
    """
    if model_name not in AVAILABLE_ONNX_MODELS:
        available_models = ", ".join(AVAILABLE_ONNX_MODELS.keys())
        raise ValueError(f"Unknown model {model_name}. Use one of [{available_models}]")

    if save_directory is None:
        save_directory = MODEL_CACHE_DIR / model_name
    elif save_directory.is_file():
        raise ValueError(f"Expected a directory, but got {save_directory}")

    save_directory.mkdir(parents=True, exist_ok=True)

    model_url, config_url = AVAILABLE_ONNX_MODELS[model_name]
    model_filename = save_directory / model_url.split("/")[-1]
    config_filename = save_directory / config_url.split("/")[-1]

    if not force_download and model_filename.is_file() and config_filename.is_file():
        logging.info(
            "Skipping download of '%s' model, already exists at %s",
            model_name,
            save_directory,
        )
        return model_filename, config_filename

    # Download the model if not present or if we want to force the download
    if force_download or not model_filename.is_file():
        logging.info("Downloading model to %s", model_filename)
        _download_with_progress(url=model_url, filename=model_filename)

    # Same for the config
    if force_download or not config_filename.is_file():
        logging.info("Downloading config to %s", config_filename)
        _download_with_progress(url=config_url, filename=config_filename)

    return model_filename, config_filename



"""
Augmentations used for training the OCR model.
"""

import albumentations as A
import cv2

BORDER_COLOR_BLACK: tuple[int, int, int] = (0, 0, 0)

TRAIN_AUGMENTATION = A.Compose(
    [
        A.ShiftScaleRotate(
            shift_limit=0.06,
            scale_limit=0.1,
            rotate_limit=9,
            border_mode=cv2.BORDER_CONSTANT,
            value=BORDER_COLOR_BLACK,
            p=1,
        ),
        A.RandomBrightnessContrast(brightness_limit=0.1, contrast_limit=0.1, p=1),
        A.MotionBlur(blur_limit=(3, 5), p=0.1),
        A.OneOf(
            [
                A.CoarseDropout(max_holes=10, max_height=4, max_width=4, p=0.3),
                A.PixelDropout(dropout_prob=0.01, p=0.2),
            ],
            p=0.7,
        ),
    ]
)
"""Training augmentations recipe."""



"""
Layer blocks used in the OCR model.
"""

from keras import regularizers
from keras.src.layers import (
    Activation,
    AveragePooling2D,
    BatchNormalization,
    Conv2D,
    MaxPooling2D,
    SeparableConv2D,
)


def block_no_bn(i, k=3, n_c=64, s=1, padding="same", activation: str = "relu"):
    x1 = Conv2D(
        kernel_size=k,
        filters=n_c,
        strides=s,
        padding=padding,
        kernel_regularizer=regularizers.l2(0.01),
        use_bias=False,
    )(i)
    x2 = Activation(activation)(x1)
    return x2, x1


def block_no_activation(i, k=3, n_c=64, s=1, padding="same"):
    x = Conv2D(
        kernel_size=k,
        filters=n_c,
        strides=s,
        padding=padding,
        kernel_regularizer=regularizers.l2(0.01),
        use_bias=False,
    )(i)
    x = BatchNormalization()(x)
    return x


def block_bn(i, k=3, n_c=64, s=1, padding="same", activation: str = "relu"):
    x1 = Conv2D(
        kernel_size=k,
        filters=n_c,
        strides=s,
        padding=padding,
        kernel_regularizer=regularizers.l2(0.01),
        use_bias=False,
    )(i)
    x2 = BatchNormalization()(x1)
    x2 = Activation(activation)(x2)
    return x2, x1


def block_bn_no_l2(i, k=3, n_c=64, s=1, padding="same", activation: str = "relu"):
    x1 = Conv2D(kernel_size=k, filters=n_c, strides=s, padding=padding, use_bias=False)(i)
    x2 = BatchNormalization()(x1)
    x2 = Activation(activation)(x2)
    return x2, x1


def block_bn_sep_conv_l2(
    i, k=3, n_c=64, s=1, padding="same", depth_multiplier=1, activation: str = "relu"
):
    l2_kernel_reg = regularizers.l2(0.01)
    x1 = SeparableConv2D(
        kernel_size=k,
        filters=n_c,
        depth_multiplier=depth_multiplier,
        strides=s,
        padding=padding,
        use_bias=False,
        depthwise_regularizer=l2_kernel_reg,
        pointwise_regularizer=l2_kernel_reg,
    )(i)
    x2 = BatchNormalization()(x1)
    x2 = Activation(activation)(x2)
    return x2, x1


def block_bn_relu6(i, k=3, n_c=64, s=1, padding="same", activation: str = "relu6"):
    x1 = Conv2D(
        kernel_size=k,
        filters=n_c,
        strides=s,
        padding=padding,
        kernel_regularizer=regularizers.l2(0.01),
        use_bias=False,
    )(i)
    x2 = BatchNormalization()(x1)
    x2 = Activation(activation)(x2)
    return x2, x1


def block_bn_relu6_no_l2(i, k=3, n_c=64, s=1, padding="same", activation: str = "relu6"):
    x1 = Conv2D(kernel_size=k, filters=n_c, strides=s, padding=padding, use_bias=False)(i)
    x2 = BatchNormalization()(x1)
    x2 = Activation(activation)(x2)
    return x2, x1


def block_average_conv_down(x, n_c, padding="same", activation: str = "relu"):
    x = AveragePooling2D(pool_size=2, strides=1, padding="valid")(x)
    x = Conv2D(
        filters=n_c,
        kernel_size=3,
        strides=2,
        padding=padding,
        kernel_regularizer=regularizers.l2(0.01),
        use_bias=False,
    )(x)
    x = BatchNormalization()(x)
    x = Activation(activation)(x)
    return x


def block_max_conv_down(x, n_c, padding="same", activation: str = "relu"):
    x = MaxPooling2D(pool_size=2, strides=1, padding="valid")(x)
    x = Conv2D(
        filters=n_c,
        kernel_size=3,
        strides=2,
        padding=padding,
        kernel_regularizer=regularizers.l2(0.01),
        use_bias=False,
    )(x)
    x = BatchNormalization()(x)
    x = Activation(activation)(x)
    return x



"""
Script for displaying an image with the OCR model predictions.
"""

import logging
import pathlib

import click
import cv2
import keras
import numpy as np

from fast_plate_ocr.train.model.config import load_config_from_yaml
from fast_plate_ocr.train.utilities import utils
from fast_plate_ocr.train.utilities.utils import postprocess_model_output

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
)


@click.command(context_settings={"max_content_width": 120})
@click.option(
    "-m",
    "--model",
    "model_path",
    required=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, path_type=pathlib.Path),
    help="Path to the saved .keras model.",
)
@click.option(
    "--config-file",
    required=True,
    type=click.Path(exists=True, file_okay=True, path_type=pathlib.Path),
    help="Path pointing to the model license plate OCR config.",
)
@click.option(
    "-d",
    "--img-dir",
    required=True,
    type=click.Path(exists=True, dir_okay=True, file_okay=False, path_type=pathlib.Path),
    help="Directory containing the images to make predictions from.",
)
@click.option(
    "-l",
    "--low-conf-thresh",
    type=float,
    default=0.35,
    show_default=True,
    help="Threshold for displaying low confidence characters.",
)
@click.option(
    "-f",
    "--filter-conf",
    type=float,
    help="Display plates that any of the plate characters are below this number.",
)
def visualize_predictions(
    model_path: pathlib.Path,
    config_file: pathlib.Path,
    img_dir: pathlib.Path,
    low_conf_thresh: float,
    filter_conf: float | None,
):
    """
    Visualize OCR model predictions on unlabeled data.
    """
    config = load_config_from_yaml(config_file)
    model = utils.load_keras_model(
        model_path, vocab_size=config.vocabulary_size, max_plate_slots=config.max_plate_slots
    )
    images = utils.load_images_from_folder(
        img_dir, width=config.img_width, height=config.img_height
    )
    for image in images:
        x = np.expand_dims(image, 0)
        prediction = model(x, training=False)
        prediction = keras.ops.stop_gradient(prediction).numpy()
        plate, probs = postprocess_model_output(
            prediction=prediction,
            alphabet=config.alphabet,
            max_plate_slots=config.max_plate_slots,
            vocab_size=config.vocabulary_size,
        )
        if not filter_conf or (filter_conf and np.any(probs < filter_conf)):
            utils.display_predictions(
                image=image, plate=plate, probs=probs, low_conf_thresh=low_conf_thresh
            )
    cv2.destroyAllWindows()


if __name__ == "__main__":
    visualize_predictions()
